/* sgets.c
**
** Readoperation which reads a string of arbitrary length and automatically
** continues after an incoming SIGCHLD signal.
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "tools.h"
#include "mrmacs.h"

#include "sgets.h"

/* sgets (stream, &cbuf, &cblen, &done)
**       - liest eine Zeile in einen dynamisch alloziierten Puffer `cbuf' ein.
**         Die augenblickliche Lnge dieses Puffers steht in `cblen'. Reicht
**         der Speicherplatz nicht mehr aus, dann wird neuer Speicherplatz fr
**         `cbuf' alloziiert, der ausreichend fr die Aufnahme der bisher ein-
**         gelesenen Daten + 128 Bytes ist. Die Adresse dieses Puffers wird in
**         `cbuf' zurckgegeben, die neue Lnge in `cblen'. Das Flag `done'
**         sagt aus, ob die Operation als ganzes erfolgreich war (done=1) oder
**         nicht (done = 0).
**         Fr `cbuf' MUSS eine Zeigervariable (Typ: char *) angegeben werden,
**         die entweder auf einen mittels `galloc' alloziierten Bereich zeigt,
**         oder NULL enthlt. In `done' mu die aktuelle Puffergre bergeben
**         werden - NICHT die Zeichenkettenlnge seines Inhalts!.
*/
char *sgets (FILE *fp, char **_base, size_t *_bsize, int *done)
{
#define base (*_base)
#define bsize (*_bsize)
    char *buf, *p;
    size_t nlen;
    /* Beschaffe die Anfangsmenge an Speicherplatz fr den Puffer */
    ifnull (base) {
	bsize = 256;
	base = (char *) galloc (__FILE__, __LINE__, bsize);
    }
    buf = base;
    /* Zunchst sollen genausoviele Zeichen eingelesen werden, wie der Puffer
    ** lang ist. Falls der Puffer bereits in einer vorherigen Leseoperation
    ** ausreichend vergrert wurde, kehrt `sgets' bereits nach der ersten
    ** Leseoperation zurck.
    */
    nlen = bsize;
    /* Den Puffer vorher richtig initialisieren! */
    *base = '\0';
    *done = 1;
    for (;;) {
REDO:
	ifnull (fgets (buf, nlen, fp)) {
	    /* Wurde die Aktion nur durch ein Signal (SIGCHLD) unterbrochen,
	    ** dann soll sie ohne Umschweife fortgesetzt werden.
	    */
	    if (errno == EINTR) { goto REDO; }
	    /* Jeder andere Fehlertyp fhrt zur Rckgabe der bisher gelesenen
	    ** Zeichen und einem Rcksetzen des Flag-Argumentes.
	    */
	    *done = 0;
	    break;
	}
	/* Bestimme das letzte eingelesene Zeichen. */
	p = &buf[strlen (buf) - 1];
	/* War das letzte gelesene Zeichen (for dem terminierenden NUL) ein
	** Zeilenendezeichen (CR beim Mac, LF unter Unix und CR/LF unter
	** DOof/Windoof), dann ist `sgets' hiermit beendet. Das
	** Flag-Argument wird gestzt und die Schleife beendet.
	*/
	if (*p == '\n' || *p == '\r') { *done = 1; break; }
	/* Sonst wird der Abstand des Zeichenkettenendes zum Anfang bestimmt
	** (Retten der Zeiger fr den anschlieenden `gresize'),
	*/
	nlen = &p[1] - buf;
	/* danach der Puffer um 128 Zeichen vergrert, */
	bsize += 128;
	base = (char *) gresize (__FILE__, __LINE__, base, bsize);
	/* Danach die aktuelle Pufferposition wiederhergestellt, */
	buf = &base[nlen];
	/* Und die maximalzahl der nun zu lesenden Zeichen auf 128 gesetzt. */
	nlen = 128;
    }
    return base;
#undef bsize
#undef base
}
